import logging
import os
import json
import itertools
import dpath.util
from checkov.common.models.consts import SUPPORTED_FILE_EXTENSIONS

logging.basicConfig(level=logging.INFO)
# define a Handler which writes INFO messages or higher to the sys.stderr
console = logging.StreamHandler()
console.setLevel(logging.INFO)
# set a format which is simpler for console use
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
# tell the handler to use this format
console.setFormatter(formatter)

checkov_results_prefix = 'checkov_results'
check_reduced_keys = (
    'check_id', 'check_result', 'resource', 'file_path',
    'file_line_range')
check_metadata_keys = ('evaluations', 'code_block')


def _is_scanned_file(file):
    file_ending = os.path.splitext(file)[1]
    return file_ending in SUPPORTED_FILE_EXTENSIONS


def _put_json_object(s3_client, json_obj, bucket, object_path):
    try:
        s3_client.put_object(Bucket=bucket, Key=object_path, Body=json.dumps(json_obj))
    except Exception as e:
        logging.error(f"failed to persist object {json_obj} into S3 bucket {bucket}\n{e}")
        raise e


def _extract_checks_metadata(report, full_repo_object_key):
    return {check.check_id: dict({k: getattr(check, k) for k in check_metadata_keys},
                                 **{'file_object_path': full_repo_object_key + check.file_path}) for check in
            list(itertools.chain(report.passed_checks, report.failed_checks, report.skipped_checks))}


def reduce_scan_reports(scan_reports):
    """
    Transform checkov reports objects into compact dictionaries
    :param scan_reports: List of checkov output reports
    :return: dictionary of
    """
    reduced_scan_reports = {}
    for report in scan_reports:
        reduced_scan_reports[report.check_type] = \
            {
                "checks": {
                    "passed_checks": [
                        {k: getattr(check, k) for k in check_reduced_keys}
                        for check in report.passed_checks],
                    "failed_checks": [
                        {k: getattr(check, k) for k in check_reduced_keys}
                        for check in report.failed_checks],
                    "skipped_checks": [
                        {k: getattr(check, k) for k in check_reduced_keys}
                        for check in report.skipped_checks]}}
    return reduced_scan_reports


def persist_checks_results(reduced_scan_reports, s3_client, bucket, full_repo_object_key):
    """
    Save reduced scan reports into bridgecrew's platform
    :return: List of checks results path of all runners
    """
    checks_results_paths = {}
    for check_type, reduced_report in reduced_scan_reports.items():
        check_result_object_path = f'{full_repo_object_key}/{checkov_results_prefix}/{check_type}/checks_results.json'
        checks_results_paths[check_type] = check_result_object_path
        _put_json_object(s3_client, reduced_report, bucket, check_result_object_path)
    return checks_results_paths


def enrich_and_persist_checks_metadata(scan_reports, s3_client, bucket, full_repo_object_key):
    """
    Save checks metadata into bridgecrew's platform
    :return:
    """
    checks_metadata_paths = {}
    for scan_report in scan_reports:
        check_type = scan_report.check_type
        checks_metadata_object = _extract_checks_metadata(scan_report, full_repo_object_key)
        checks_metadata_object_path = f'{full_repo_object_key}/{checkov_results_prefix}/{check_type}/checks_metadata.json'
        dpath.new(checks_metadata_paths, f"{check_type}/checks_metadata_path", checks_metadata_object_path)
        _put_json_object(s3_client, checks_metadata_object, bucket, checks_metadata_object_path)
    return checks_metadata_paths



from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.cloudformation.checks.resource.base_resource_check import BaseResourceCheck

PORT = 22


class SecurityGroupUnrestrictedIngress22(BaseResourceCheck):
    def __init__(self):
        name = "Ensure no security groups allow ingress from 0.0.0.0:0 to port %d" % PORT
        id = "CKV_AWS_24"
        supported_resources = ['AWS::EC2::SecurityGroup', 'AWS::EC2::SecurityGroupIngress']
        categories = [CheckCategories.NETWORKING]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        """
            Looks for configuration at security group ingress rules :
            https://www.terraform.io/docs/providers/aws/r/security_group.html
        :param conf: aws_security_group configuration
        :return: <CheckResult>
        """
        rules = []
        if conf['Type'] == 'AWS::EC2::SecurityGroup':
            if 'Properties' in conf.keys():
                if 'SecurityGroupIngress' in conf['Properties'].keys():
                    rules = conf['Properties']['SecurityGroupIngress']
        elif conf['Type'] == 'AWS::EC2::SecurityGroupIngress':
            if 'Properties' in conf.keys():
                rules = []
                rules.append(conf['Properties'])

        for rule in rules:
            if rule.__contains__('FromPort') and rule.__contains__('ToPort'):
                if isinstance(rule['FromPort'], int) and isinstance(rule['ToPort'], int):
                    if int(rule['FromPort']) == int(PORT) and int(rule['ToPort']) == int(PORT):
                        if 'CidrIp' in rule.keys() and rule['CidrIp'] == '0.0.0.0/0':
                            return CheckResult.FAILED
                        elif 'CidrIpv6' in rule.keys() and rule['CidrIpv6'] == '::/0':
                            return CheckResult.FAILED
        return CheckResult.PASSED


check = SecurityGroupUnrestrictedIngress22()



from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck
from checkov.common.models.enums import CheckResult, CheckCategories


class GKEAliasIpEnabled(BaseResourceCheck):
    def __init__(self):
        name = "Ensure Kubernetes Cluster is created with Alias IP ranges enabled"
        id = "CKV_GCP_23"
        supported_resources = ['google_container_cluster']
        categories = [CheckCategories.KUBERNETES]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        """
            Looks for password configuration at azure_instance:
            https://www.terraform.io/docs/providers/google/r/compute_ssl_policy.html
        :param conf: google_compute_ssl_policy configuration
        :return: <CheckResult>
        """
        if conf.get('ip_allocation_policy'):
            return CheckResult.PASSED
        return CheckResult.FAILED


check = GKEAliasIpEnabled()



from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck


class AKSRbacEnabled(BaseResourceCheck):
    def __init__(self):
        name = "Ensure RBAC is enabled on AKS clusters"
        id = "CKV_AZURE_5"
        supported_resources = ['azurerm_kubernetes_cluster']
        categories = [CheckCategories.KUBERNETES]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        if 'role_based_access_control' not in conf or conf['role_based_access_control'][0]['enabled'][0]:
            return CheckResult.PASSED
        return CheckResult.FAILED


check = AKSRbacEnabled()



import logging
import operator
import os
from functools import reduce

from checkov.cloudformation.checks.resource.registry import resource_registry
from checkov.cloudformation.parser import parse
from checkov.common.output.record import Record
from checkov.common.output.report import Report
from checkov.common.runners.base_runner import BaseRunner
from checkov.runner_filter import RunnerFilter

CF_POSSIBLE_ENDINGS = [".yml", ".yaml", ".json", ".template"]


class Runner(BaseRunner):
    check_type = "cloudformation"

    def run(self, root_folder, external_checks_dir=None, files=None, runner_filter=RunnerFilter()):
        report = Report(self.check_type)
        definitions = {}
        definitions_raw = {}
        parsing_errors = {}
        files_list = []
        if external_checks_dir:
            for directory in external_checks_dir:
                resource_registry.load_external_checks(directory)

        if files:
            for file in files:
                (definitions[file], definitions_raw[file]) = parse(file)

        if root_folder:
            for root, d_names, f_names in os.walk(root_folder):
                for file in f_names:
                    file_ending = os.path.splitext(file)[1]
                    if file_ending in CF_POSSIBLE_ENDINGS:
                        files_list.append(os.path.join(root, file))

            for file in files_list:
                relative_file_path = f'/{os.path.relpath(file, os.path.commonprefix((root_folder, file)))}'
                (definitions[relative_file_path], definitions_raw[relative_file_path]) = parse(file)

        # Filter out empty files that have not been parsed successfully, and filter out non-CF template files
        definitions = {k: v for k, v in definitions.items() if v and v.__contains__("Resources")}
        definitions_raw = {k: v for k, v in definitions_raw.items() if k in definitions.keys()}

        for cf_file in definitions.keys():
            if not 'Resources' in definitions[cf_file].keys():
                continue
            logging.debug("Template Dump for {}: {}".format(cf_file, definitions[cf_file], indent=2))

            # Get Parameter Defaults - Locate Refs in Template
            refs = []
            refs.extend(self._search_deep_keys('Ref', definitions[cf_file], []))

            for ref in refs:
                refname = ref.pop()
                ref.pop()  # Get rid of the 'Ref' dict key

                if 'Parameters' in definitions[cf_file].keys() and refname in definitions[cf_file][
                    'Parameters'].keys():
                    # TODO refactor into evaluations
                    if 'Default' in definitions[cf_file]['Parameters'][refname].keys():
                        logging.debug(
                            "Replacing Ref {} in file {} with default parameter value: {}".format(refname, cf_file,
                                                                                                  definitions[
                                                                                                      cf_file][
                                                                                                      'Parameters'][
                                                                                                      refname][
                                                                                                      'Default']))
                        _set_in_dict(definitions[cf_file], ref,
                                     definitions[cf_file]['Parameters'][refname]['Default'])

                        ## TODO - Add Variable Eval Message for Output
                        # Output in Checkov looks like this:
                        # Variable versioning (of /.) evaluated to value "True" in expression: enabled = ${var.versioning}

            for resource_name, resource in definitions[cf_file]['Resources'].items():
                if resource_name == '__startline__' or resource_name == '__endline__':
                    continue
                resource_id = f"{resource['Type']}.{resource_name}"

                ## TODO - Evaluate skipped_checks
                skipped_checks = {}

                results = resource_registry.scan(cf_file, {resource_name: resource}, skipped_checks,runner_filter)
                # TODO refactor into context parsing
                find_lines_result_list = list(find_lines(resource, '__startline__'))
                if len(find_lines_result_list) >= 1:
                    start_line = min(find_lines_result_list)
                    end_line = max(list(find_lines(resource, '__endline__')))

                    entity_lines_range = [start_line, end_line - 1]

                    entity_code_lines = definitions_raw[cf_file][start_line - 1: end_line - 1]

                    # TODO - Variable Eval Message!
                    variable_evaluations = {}

                    for check, check_result in results.items():
                        ### TODO - Need to get entity_code_lines and entity_lines_range
                        record = Record(check_id=check.id, check_name=check.name, check_result=check_result,
                                        code_block=entity_code_lines, file_path=cf_file,
                                        file_line_range=entity_lines_range,
                                        resource=resource_id, evaluations=variable_evaluations,
                                        check_class=check.__class__.__module__)
                        report.add_record(record=record)
        return report

    def _search_deep_keys(self, search_text, cfn_dict, path):
        """Search deep for keys and get their values"""
        keys = []
        if isinstance(cfn_dict, dict):
            for key in cfn_dict:
                pathprop = path[:]
                pathprop.append(key)
                if key == search_text:
                    pathprop.append(cfn_dict[key])
                    keys.append(pathprop)
                    # pop the last element off for nesting of found elements for
                    # dict and list checks
                    pathprop = pathprop[:-1]
                if isinstance(cfn_dict[key], dict):
                    keys.extend(self._search_deep_keys(search_text, cfn_dict[key], pathprop))
                elif isinstance(cfn_dict[key], list):
                    for index, item in enumerate(cfn_dict[key]):
                        pathproparr = pathprop[:]
                        pathproparr.append(index)
                        keys.extend(self._search_deep_keys(search_text, item, pathproparr))
        elif isinstance(cfn_dict, list):
            for index, item in enumerate(cfn_dict):
                pathprop = path[:]
                pathprop.append(index)
                keys.extend(self._search_deep_keys(search_text, item, pathprop))

        return keys


def _get_from_dict(data_dict, map_list):
    return reduce(operator.getitem, map_list, data_dict)


def _set_in_dict(data_dict, map_list, value):
    _get_from_dict(data_dict, map_list[:-1])[map_list[-1]] = value


def find_lines(node, kv):
    if isinstance(node, list):
        for i in node:
            for x in find_lines(i, kv):
                yield x
    elif isinstance(node, dict):
        if kv in node:
            yield node[kv]
        for j in node.values():
            for x in find_lines(j, kv):
                yield x



import re

from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck
from checkov.terraform.checks.utils.consts import access_key_pattern, secret_key_pattern


class EC2Credentials(BaseResourceCheck):

    def __init__(self):
        name = "Ensure no hard coded AWS access key and and secret key exists in EC2 user data"
        id = "CKV_AWS_46"
        supported_resources = ['aws_instance']
        categories = [CheckCategories.SECRETS]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        if 'user_data' in conf.keys():
            user_data = conf['user_data'][0]
            if isinstance(user_data, str):
                if re.match(".*{}.*".format(access_key_pattern), conf['user_data'][0], re.DOTALL) or re.match(
                        ".*{}.*".format(secret_key_pattern), conf['user_data'][0], re.DOTALL):
                    return CheckResult.FAILED
        return CheckResult.PASSED


check = EC2Credentials()



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.kubernetes.base_spec_check import BaseK8Check


class DropCapabilitiesPSP(BaseK8Check):

    def __init__(self):
        # CIS-1.3 1.7.7
        # CIS-1.5 5.2.7
        name = "Do not admit containers with the NET_RAW capability"
        # Location: PodSecurityPolicy.spec.requiredDropCapabilities
        id = "CKV_K8S_7"
        supported_kind = ['PodSecurityPolicy']
        categories = [CheckCategories.KUBERNETES]
        super().__init__(name=name, id=id, categories=categories, supported_entities=supported_kind)

    def get_resource_id(self, conf):
        if "metadata" in conf:
            if "name" in conf["metadata"]:
                return 'PodSecurityPolicy.{}'.format(conf["metadata"]["name"])
        return 'PodSecurityPolicy.spec.requiredDropCapabilities'

    def scan_spec_conf(self, conf):
        if "spec" in conf:
            if "requiredDropCapabilities" in conf["spec"]:
                if "ALL" in conf["spec"]["requiredDropCapabilities"] or "NET_RAW" in conf["spec"][
                    "requiredDropCapabilities"]:
                    return CheckResult.PASSED
        return CheckResult.FAILED


check = DropCapabilitiesPSP()



from checkov.terraform.checks.resource.base_resource_check import BaseResourceCheck
from checkov.common.models.enums import CheckResult, CheckCategories


class GKEPodSecurityPolicyEnabled(BaseResourceCheck):
    def __init__(self):
        name = "Ensure Kubernetes Cluster is created with Private cluster enabled"
        id = "CKV_GCP_25"
        supported_resources = ['google_container_cluster']
        categories = [CheckCategories.KUBERNETES]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        if conf.get('private_cluster_config'):
            return CheckResult.PASSED
        return CheckResult.FAILED


check = GKEPodSecurityPolicyEnabled()
